@@ -1,7 +1,7 @@
 /* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
 /* GeoIP.c
  *
- * Copyright (C) 2006 MaxMind LLC
+ * Copyright (C) 2016 MaxMind, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -872,7 +872,7 @@ char * _GeoIP_iso_8859_1__utf8(const char * iso)
     if (p) {
         while ( ( c = *iso++ ) ) {
             if (c < 0) {
-                k = (char) 0xc2;
+                k = (char)0xc2;
                 if (c >= -64) {
                     k++;
                 }
@@ -959,7 +959,7 @@ static void _setup_segments(GeoIP * gi)
     int i, j, segment_record_length;
     unsigned char delim[3];
     unsigned char buf[LARGE_SEGMENT_RECORD_LENGTH];
-
+    off_t offset = gi->size - 3;
     int fno = fileno(gi->GeoIPDatabase);
 
     gi->databaseSegments = NULL;
@@ -967,17 +967,17 @@ static void _setup_segments(GeoIP * gi)
     /* default to GeoIP Country Edition */
     gi->databaseType = GEOIP_COUNTRY_EDITION;
     gi->record_length = STANDARD_RECORD_LENGTH;
-    if (-1 == lseek(fno, -3l, SEEK_END)) {
-        return;
-    }
+
     for (i = 0; i < STRUCTURE_INFO_MAX_SIZE; i++) {
-        if (read(fno, delim, 3) != 3) {
+        if (pread(fno, delim, 3, offset) != 3) {
             return;
         }
+        offset += 3;
         if (delim[0] == 255 && delim[1] == 255 && delim[2] == 255) {
-            if (read(fno, &gi->databaseType, 1) != 1) {
+            if (pread(fno, &gi->databaseType, 1, offset) != 1) {
                 return;
             }
+            offset++;
             if (gi->databaseType >= 106) {
                 /* backwards compatibility with databases from April 2003 and earlier */
                 gi->databaseType -= 105;
@@ -985,14 +985,14 @@ static void _setup_segments(GeoIP * gi)
 
             if (gi->databaseType == GEOIP_REGION_EDITION_REV0) {
                 /* Region Edition, pre June 2003 */
-                gi->databaseSegments = malloc(sizeof(int));
+                gi->databaseSegments = malloc(sizeof(unsigned int));
                 if (gi->databaseSegments == NULL) {
                     return;
                 }
                 gi->databaseSegments[0] = STATE_BEGIN_REV0;
             } else if (gi->databaseType == GEOIP_REGION_EDITION_REV1) {
                 /* Region Edition, post June 2003 */
-                gi->databaseSegments = malloc(sizeof(int));
+                gi->databaseSegments = malloc(sizeof(unsigned int));
                 if (gi->databaseSegments == NULL) {
                     return;
                 }
@@ -1024,19 +1024,20 @@ static void _setup_segments(GeoIP * gi)
                        gi->databaseType == GEOIP_POSTALCONF_EDITION
                        ) {
                 /* City/Org Editions have two segments, read offset of second segment */
-                gi->databaseSegments = malloc(sizeof(int));
+                gi->databaseSegments = malloc(sizeof(unsigned int));
                 if (gi->databaseSegments == NULL) {
                     return;
                 }
                 gi->databaseSegments[0] = 0;
                 segment_record_length = SEGMENT_RECORD_LENGTH;
 
-                if (read(fno, buf,
-                         segment_record_length) != segment_record_length) {
+                if (pread(fno, buf, segment_record_length, offset)
+                    != segment_record_length) {
                     free(gi->databaseSegments);
                     gi->databaseSegments = NULL;
                     return;
                 }
+
                 for (j = 0; j < segment_record_length; j++) {
                     gi->databaseSegments[0] += (buf[j] << (j * 8));
                 }
@@ -1053,7 +1054,8 @@ static void _setup_segments(GeoIP * gi)
             }
             break;
         } else {
-            if (-1 == lseek(fno, -4l, SEEK_CUR)) {
+            offset -= 4;
+            if (offset < 0) {
                 gi->databaseSegments = NULL;
                 return;
             }
@@ -1063,20 +1065,19 @@ static void _setup_segments(GeoIP * gi)
         gi->databaseType == GEOIP_PROXY_EDITION ||
         gi->databaseType == GEOIP_NETSPEED_EDITION ||
         gi->databaseType == GEOIP_COUNTRY_EDITION_V6) {
-        gi->databaseSegments = malloc(sizeof(int));
+        gi->databaseSegments = malloc(sizeof(unsigned int));
         if (gi->databaseSegments == NULL) {
             return;
         }
         gi->databaseSegments[0] = COUNTRY_BEGIN;
     }else if (gi->databaseType == GEOIP_LARGE_COUNTRY_EDITION ||
               gi->databaseType == GEOIP_LARGE_COUNTRY_EDITION_V6) {
-        gi->databaseSegments = malloc(sizeof(int));
+        gi->databaseSegments = malloc(sizeof(unsigned int));
         if (gi->databaseSegments == NULL) {
             return;
         }
         gi->databaseSegments[0] = LARGE_COUNTRY_BEGIN;
     }
-
 }
 
 static
@@ -1510,7 +1511,7 @@ GeoIP * GeoIP_open(const char * filename, int flags)
     GeoIP * gi;
     size_t len;
 
-    gi = (GeoIP *)malloc(sizeof(GeoIP));
+    gi = (GeoIP *)calloc(1, sizeof(GeoIP));
     if (gi == NULL) {
         return NULL;
     }
@@ -1524,15 +1525,13 @@ GeoIP * GeoIP_open(const char * filename, int flags)
     gi->GeoIPDatabase = fopen(filename, "rb");
     if (gi->GeoIPDatabase == NULL) {
         DEBUG_MSGF(flags, "Error Opening file %s\n", filename);
-        free(gi->file_path);
-        free(gi);
+        GeoIP_delete(gi);
         return NULL;
     }
 
     if (fstat(fileno(gi->GeoIPDatabase), &buf) == -1) {
         DEBUG_MSGF(flags, "Error stating file %s\n", filename);
-        free(gi->file_path);
-        free(gi);
+        GeoIP_delete(gi);
         return NULL;
     }
 
@@ -1548,8 +1547,7 @@ GeoIP * GeoIP_open(const char * filename, int flags)
                          gi->GeoIPDatabase), 0);
             if (gi->cache == MAP_FAILED) {
                 DEBUG_MSGF(flags, "Error mmaping file %s\n", filename);
-                free(gi->file_path);
-                free(gi);
+                GeoIP_delete(gi);
                 return NULL;
             }
 #endif
@@ -1561,9 +1559,7 @@ GeoIP * GeoIP_open(const char * filename, int flags)
                 if (pread(fileno(gi->GeoIPDatabase), gi->cache, buf.st_size,
                           0) != (ssize_t)buf.st_size) {
                     DEBUG_MSGF(flags, "Error reading file %s\n", filename);
-                    free(gi->cache);
-                    free(gi->file_path);
-                    free(gi);
+                    GeoIP_delete(gi);
                     return NULL;
                 }
             }
@@ -1572,8 +1568,7 @@ GeoIP * GeoIP_open(const char * filename, int flags)
         if (flags & GEOIP_CHECK_CACHE) {
             if (fstat(fileno(gi->GeoIPDatabase), &buf) == -1) {
                 DEBUG_MSGF(flags, "Error stating file %s\n", filename);
-                free(gi->file_path);
-                free(gi);
+                GeoIP_delete(gi);
                 return NULL;
             }
             gi->mtime = buf.st_mtime;
@@ -1584,23 +1579,19 @@ GeoIP * GeoIP_open(const char * filename, int flags)
     gi->charset = GEOIP_CHARSET_ISO_8859_1;
     gi->ext_flags = 1U << GEOIP_TEREDO_BIT;
     _setup_segments(gi);
+    if (gi->databaseSegments == NULL) {
+        DEBUG_MSGF(gi->flags, "Error reading file %s -- corrupt\n",
+                   gi->file_path);
+        GeoIP_delete(gi);
+        return NULL;
+    }
+
 
     idx_size = get_index_size(gi, &buf);
 
     if (idx_size < 0) {
         DEBUG_MSGF(gi->flags, "Error file %s -- corrupt\n", gi->file_path);
-        if (flags & GEOIP_MEMORY_CACHE) {
-            free(gi->cache);
-        }
-#if !defined(_WIN32)
-        else if (gi->cache && (flags & GEOIP_MMAP_CACHE)) {
-            /* MMAP is only avail on UNIX */
-            munmap(gi->cache, gi->size);
-            gi->cache = NULL;
-        }
-#endif
-        free(gi->file_path);
-        free(gi);
+        GeoIP_delete(gi);
         return NULL;
     }
 
@@ -1611,9 +1602,7 @@ GeoIP * GeoIP_open(const char * filename, int flags)
             if (pread(fileno(gi->GeoIPDatabase), gi->index_cache, idx_size,
                       0) != idx_size) {
                 DEBUG_MSGF(gi->flags, "Error reading file %s\n", filename);
-                free(gi->databaseSegments);
-                free(gi->index_cache);
-                free(gi);
+                GeoIP_delete(gi);
                 return NULL;
             }
         }
@@ -1646,15 +1635,9 @@ void GeoIP_delete(GeoIP *gi)
         }
         gi->cache = NULL;
     }
-    if (gi->index_cache != NULL) {
-        free(gi->index_cache);
-    }
-    if (gi->file_path != NULL) {
-        free(gi->file_path);
-    }
-    if (gi->databaseSegments != NULL) {
-        free(gi->databaseSegments);
-    }
+    free(gi->index_cache);
+    free(gi->file_path);
+    free(gi->databaseSegments);
     free(gi);
 }
 
@@ -2026,8 +2009,8 @@ char *GeoIP_database_info(GeoIP * gi)
     int i;
     unsigned char buf[3];
     char *retval;
-    int hasStructureInfo = 0;
-
+    int has_structure_info = 0;
+    off_t offset = gi->size - 3;
     int fno;
 
     if (gi == NULL) {
@@ -2037,50 +2020,53 @@ char *GeoIP_database_info(GeoIP * gi)
     fno = fileno(gi->GeoIPDatabase);
 
     _check_mtime(gi);
-    if (-1 == lseek(fno, -3l, SEEK_END)) {
-        return NULL;
-    }
 
     /* first get past the database structure information */
     for (i = 0; i < STRUCTURE_INFO_MAX_SIZE; i++) {
-        if (read(fno, buf, 3 ) != 3) {
+        if (pread(fno, buf, 3, offset) != 3) {
             return NULL;
         }
+        offset += 3;
         if (buf[0] == 255 && buf[1] == 255 && buf[2] == 255) {
-            hasStructureInfo = 1;
+            has_structure_info = 1;
             break;
         }
-        if (-1 == lseek(fno, -4l, SEEK_CUR)) {
+        offset -= 4;
+        if (offset < 0) {
             return NULL;
         }
     }
-    if (hasStructureInfo == 1) {
-        if (-1 == lseek(fno, -6l, SEEK_CUR)) {
+    if (has_structure_info) {
+        offset -= 6;
+        if (offset < 0) {
             return NULL;
         }
     } else {
         /* no structure info, must be pre Sep 2002 database, go back to end */
-        if (-1 == lseek(fno, -3l, SEEK_END)) {
+        offset = gi->size - 3;
+        if (offset < 0) {
             return NULL;
         }
     }
 
     for (i = 0; i < DATABASE_INFO_MAX_SIZE; i++) {
-        if (read(fno, buf, 3 ) != 3) {
+        if (pread(fno, buf, 3, offset) != 3) {
             return NULL;
         }
+        offset += 3;
         if (buf[0] == 0 && buf[1] == 0 && buf[2] == 0) {
             retval = malloc(sizeof(char) * (i + 1));
             if (retval == NULL) {
                 return NULL;
             }
-            if (read(fno, retval, i) != i) {
+            if (pread(fno, retval, i, offset) != i) {
                 return NULL;
             }
             retval[i] = '\0';
             return retval;
         }
-        if (-1 == lseek(fno, -4l, SEEK_CUR)) {
+        offset -= 4;
+        if (offset < 0) {
             return NULL;
         }
     }
